#include "i2-configcomponent.h"

using namespace icinga;
using std::dynamic_pointer_cast;

IcingaApplication::RefType ConfigComponent::GetIcingaApplication(void)
{
	return dynamic_pointer_cast<IcingaApplication>(GetApplication());
}

string ConfigComponent::GetName(void)
{
	return "configcomponent";
}

void ConfigComponent::Start(void)
{
	IcingaApplication::RefType icingaApp = GetIcingaApplication();

	if (icingaApp.get() == NULL)
		throw exception(/*"Component loaded by incompatible application."*/);

	ConnectionManager::RefType connectionManager = icingaApp->GetConnectionManager();
	ConfigHive::RefType configHive = icingaApp->GetConfigHive();

	if (GetConfig()->GetPropertyInteger("configSource") != 0) {
		connectionManager->RegisterMethod("config::FetchObjects", bind_weak(&ConfigComponent::FetchObjectsHandler, shared_from_this()));

		configHive->OnObjectCreated.bind(bind_weak(&ConfigComponent::LocalObjectCreatedHandler, shared_from_this()));
		configHive->OnObjectRemoved.bind(bind_weak(&ConfigComponent::LocalObjectRemovedHandler, shared_from_this()));
		configHive->OnPropertyChanged.bind(bind_weak(&ConfigComponent::LocalPropertyChangedHandler, shared_from_this()));
	}

	connectionManager->RegisterMethod("config::ObjectCreated", bind_weak(&ConfigComponent::RemoteObjectCreatedHandler, shared_from_this()));
	connectionManager->RegisterMethod("config::ObjectRemoved", bind_weak(&ConfigComponent::RemoteObjectRemovedHandler, shared_from_this()));
	connectionManager->RegisterMethod("config::PropertyChanged", bind_weak(&ConfigComponent::RemotePropertyChangedHandler, shared_from_this()));
}

void ConfigComponent::Stop(void)
{
	// TODO: implement
}

JsonRpcMessage::RefType ConfigComponent::MakeObjectMessage(const ConfigObject::RefType& object, string method, bool includeProperties)
{
	JsonRpcMessage::RefType msg = new_object<JsonRpcMessage>();
	msg->SetVersion("2.0");
	msg->SetMethod(method);
	cJSON *params = msg->GetParams();

	string name = object->GetName();
	cJSON_AddStringToObject(params, "name", name.c_str());

	string type = object->GetType();
	cJSON_AddStringToObject(params, "type", type.c_str());

	if (includeProperties) {
		for (ConfigObject::ParameterIterator pi = object->Properties.begin(); pi != object->Properties.end(); pi++) {
			cJSON_AddStringToObject(params, pi->first.c_str(), pi->second.c_str());
		}
	}

	return msg;
}

int ConfigComponent::FetchObjectsHandler(NewMessageEventArgs::RefType ea)
{
	JsonRpcClient::RefType client = static_pointer_cast<JsonRpcClient>(ea->Source);
	ConfigHive::RefType configHive = GetIcingaApplication()->GetConfigHive();

	for (ConfigHive::TypeIterator ti = configHive->Objects.begin(); ti != configHive->Objects.end(); ti++) {
		for (ConfigHive::ObjectIterator oi = ti->second.begin(); oi != ti->second.end(); oi++) {
			JsonRpcMessage::RefType msg = MakeObjectMessage(oi->second, "config::ObjectCreated", true);
			client->SendMessage(msg);
		}
	}

	return 0;
}

int ConfigComponent::LocalObjectCreatedHandler(ConfigHiveEventArgs::RefType ea)
{
	ConnectionManager::RefType connectionManager = GetIcingaApplication()->GetConnectionManager();
	connectionManager->SendMessage(MakeObjectMessage(ea->ConfigObject, "config::ObjectCreated", true));

	return 0;
}

int ConfigComponent::LocalObjectRemovedHandler(ConfigHiveEventArgs::RefType ea)
{
	ConnectionManager::RefType connectionManager = GetIcingaApplication()->GetConnectionManager();
	connectionManager->SendMessage(MakeObjectMessage(ea->ConfigObject, "config::ObjectRemoved", false));

	return 0;
}

int ConfigComponent::LocalPropertyChangedHandler(ConfigHiveEventArgs::RefType ea)
{
	JsonRpcMessage::RefType msg = MakeObjectMessage(ea->ConfigObject, "config::ObjectRemoved", false);
	cJSON *params = msg->GetParams();
	cJSON_AddStringToObject(params, "property", ea->Property.c_str());
	string value = ea->ConfigObject->GetProperty(ea->Property);
	cJSON_AddStringToObject(params, "value", value.c_str());

	ConnectionManager::RefType connectionManager = GetIcingaApplication()->GetConnectionManager();
	connectionManager->SendMessage(msg);

	return 0;
}

int ConfigComponent::RemoteObjectCreatedHandler(NewMessageEventArgs::RefType ea)
{
	JsonRpcMessage::RefType message = ea->Message;

	// TODO: update hive
	return 0;
}

int ConfigComponent::RemoteObjectRemovedHandler(NewMessageEventArgs::RefType ea)
{
	JsonRpcMessage::RefType message = ea->Message;
	string name, type;
	
	if (!message->GetParamString("name", &name))
		return 0;

	if (!message->GetParamString("type", &type))
		return 0;

	ConfigHive::RefType configHive = GetIcingaApplication()->GetConfigHive();
	ConfigObject::RefType object = configHive->GetObject(type, name);

	if (object.get() == NULL)
		return 0;

	configHive->RemoveObject(object);

	return 0;
}

int ConfigComponent::RemotePropertyChangedHandler(NewMessageEventArgs::RefType ea)
{
	JsonRpcMessage::RefType message = ea->Message;
	string name, type, property, value;

	if (!message->GetParamString("name", &name))
		return 0;

	if (!message->GetParamString("type", &type))
		return 0;

	if (!message->GetParamString("property", &property))
		return 0;

	if (!message->GetParamString("value", &value))
		return 0;

	ConfigHive::RefType configHive = GetIcingaApplication()->GetConfigHive();
	ConfigObject::RefType object = configHive->GetObject(type, name);

	if (object.get() == NULL)
		return 0;

	object->SetProperty(property, value);

	return 0;
}


EXPORT_COMPONENT(ConfigComponent);
